/*
 *  GetMail by Davide Libenzi ( POP3 Internet mail retriever )
 *  Copyright (C) 2000,..., 2003  Davide Libenzi
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Davide Libenzi <davidel@xmailserver.org>
 *
 */



#include <winsock2.h>
#include <mswsock.h>
#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include <ctype.h>
#include <limits.h>
#include "sockio.h"
#include "BuffSock.h"





#define GM_TIMEOUT              60
#define RECV_BUFFER_SIZE        2048

#define ResultIsOK(s)           (strnicmp((s), "+OK", 3) == 0)





static BOOL     GM_GetResponse(BSOCK_HANDLE hBSock, int iTimeout, char *pszBuffer = NULL, int iMaxBuffer = 0);
static BSOCK_HANDLE GM_Login(const char *pszServer, int iPort, const char *pszUserName,
                        const char *pszPassword, int iTimeout);
static void     GM_Logout(BSOCK_HANDLE hBSock, int iTimeout);
static BOOL     GM_ExecCommand(BSOCK_HANDLE hBSock, const char *pszCommand, int iTimeout,
                        char *pszResponse = NULL, int iMaxResponse = 0);
static BOOL     GM_DoStat(BSOCK_HANDLE hBSock, long &lNumMessages, long &lBytesCount, int iTimeout);
static BOOL     GM_ExecResCommand(BSOCK_HANDLE hBSock, const char *pszRespFileName,
                        const char *pszCommand, int iTimeout);






//
// Funzione GM_GETRESPONSE
//
// @func static BOOL | GM_GetResponse | DESCRIZIONE
//
// @parm BSOCK_HANDLE | hBSock | DESCRIZIONE
// @parm int | iTimeout | DESCRIZIONE
// @parm char | *pszBuffer | DESCRIZIONE
// @parm int | iMaxBuffer | DESCRIZIONE
//
// @rdesc DESCRIZIONE_VALORE_RITORNO
//
// MadeBy : Davide Libenzi
//
static BOOL     GM_GetResponse(BSOCK_HANDLE hBSock, int iTimeout, char *pszBuffer, int iMaxBuffer)
{

    char            szRespBuffer[RECV_BUFFER_SIZE] = "";

    if (BSckGetString(hBSock, szRespBuffer, sizeof(szRespBuffer) - 1, iTimeout) == NULL)
        return (FALSE);

    if (pszBuffer != NULL)
    {
        int             iRespLen = strlen(szRespBuffer),
                        iCopyLen = min(iRespLen, iMaxBuffer);

        strncpy(pszBuffer, szRespBuffer, iRespLen);
        pszBuffer[iRespLen] = '\0';
    }

    return ((ResultIsOK(szRespBuffer)) ? TRUE : FALSE);

}




//
// Funzione GM_LOGIN
//
// @func static BSOCK_HANDLE | GM_Login | DESCRIZIONE
//
// @parm const char | *pszServer | DESCRIZIONE
// @parm int | iPort | DESCRIZIONE
// @parm const char | *pszUserName | DESCRIZIONE
// @parm const char | *pszPassword | DESCRIZIONE
// @parm int | iTimeout | DESCRIZIONE
//
// @rdesc DESCRIZIONE_VALORE_RITORNO
//
// MadeBy : Davide Libenzi
//
static BSOCK_HANDLE GM_Login(const char *pszServer, int iPort, const char *pszUserName,
                        const char *pszPassword, int iTimeout)
{

    unsigned long   ulNetAddr = INADDR_NONE;
    struct hostent *pHostEnt = NULL;

    if (((ulNetAddr = inet_addr(pszServer)) == INADDR_NONE) &&
            ((pHostEnt = gethostbyname(pszServer)) != NULL))
        ulNetAddr = *((unsigned long int *) pHostEnt->h_addr_list[0]);

    if (ulNetAddr == INADDR_NONE)
        return (INVALID_BSOCK_HANDLE);

    SOCKET          SockFD = SvrCreateSocket(AF_INET, SOCK_STREAM, 0);

    if (SockFD == INVALID_SOCKET)
        return (INVALID_BSOCK_HANDLE);

    struct sockaddr_in SvrAddr;

    ZeroMemory(&SvrAddr, sizeof(SvrAddr));
    SvrAddr.sin_family = AF_INET;
    SvrAddr.sin_addr.s_addr = ulNetAddr;
    SvrAddr.sin_port = htons((u_short) iPort);

    if (SvrConnectTimeout(SockFD, (struct sockaddr *) & SvrAddr, sizeof(SvrAddr),
                    iTimeout) < 0)
    {
        SvrCloseSocket(SockFD);
        return (INVALID_BSOCK_HANDLE);
    }

    BSOCK_HANDLE    hBSock = BSckAttach(SockFD);

    if (hBSock == INVALID_BSOCK_HANDLE)
    {
        SvrCloseSocket(SockFD);
        return (INVALID_BSOCK_HANDLE);
    }

    if (!GM_GetResponse(hBSock, iTimeout))
    {
        GM_Logout(hBSock, iTimeout);
        return (INVALID_BSOCK_HANDLE);
    }

    char            szBuffer[RECV_BUFFER_SIZE] = "";

    sprintf(szBuffer, "USER %s", pszUserName);

    if (BSckSendString(hBSock, szBuffer, iTimeout) < 0)
    {
        GM_Logout(hBSock, iTimeout);
        return (INVALID_BSOCK_HANDLE);
    }

    if (!GM_GetResponse(hBSock, iTimeout))
    {
        GM_Logout(hBSock, iTimeout);
        return (INVALID_BSOCK_HANDLE);
    }

    sprintf(szBuffer, "PASS %s", pszPassword);

    if (BSckSendString(hBSock, szBuffer, iTimeout) < 0)
    {
        GM_Logout(hBSock, iTimeout);
        return (INVALID_BSOCK_HANDLE);
    }

    if (!GM_GetResponse(hBSock, iTimeout))
    {
        GM_Logout(hBSock, iTimeout);
        return (INVALID_BSOCK_HANDLE);
    }

    return (hBSock);

}





//
// Funzione GM_LOGOUT
//
// @func static void | GM_Logout | DESCRIZIONE
//
// @parm BSOCK_HANDLE | hBSock | DESCRIZIONE
// @parm int | iTimeout | DESCRIZIONE
//
// MadeBy : Davide Libenzi
//
static void     GM_Logout(BSOCK_HANDLE hBSock, int iTimeout)
{

    BSckSendString(hBSock, "QUIT", iTimeout);

    GM_GetResponse(hBSock, iTimeout);

    BSckDetach(hBSock, 1);

}





//
// Funzione GM_EXECCOMMAND
//
// @func static BOOL | GM_ExecCommand | DESCRIZIONE
//
// @parm BSOCK_HANDLE | hBSock | DESCRIZIONE
// @parm const char | *pszCommand | DESCRIZIONE
// @parm int | iTimeout | DESCRIZIONE
// @parm char | *pszResponse | DESCRIZIONE
// @parm int | iMaxResponse | DESCRIZIONE
//
// @rdesc DESCRIZIONE_VALORE_RITORNO
//
// MadeBy : Davide Libenzi
//
static BOOL     GM_ExecCommand(BSOCK_HANDLE hBSock, const char *pszCommand, int iTimeout,
                        char *pszResponse, int iMaxResponse)
{

    if (BSckSendString(hBSock, (char *) pszCommand, iTimeout) < 0)
        return (FALSE);

    return (GM_GetResponse(hBSock, iTimeout, pszResponse, iMaxResponse));

}





//
// Funzione GM_DOSTAT
//
// @func static BOOL | GM_DoStat | DESCRIZIONE
//
// @parm BSOCK_HANDLE | hBSock | DESCRIZIONE
// @parm long | &lNumMessages | DESCRIZIONE
// @parm long | &lBytesCount | DESCRIZIONE
// @parm int | iTimeout | DESCRIZIONE
//
// @rdesc DESCRIZIONE_VALORE_RITORNO
//
// MadeBy : Davide Libenzi
//
static BOOL     GM_DoStat(BSOCK_HANDLE hBSock, long &lNumMessages, long &lBytesCount, int iTimeout)
{

    char            szResponse[1024] = "";

    if (!GM_ExecCommand(hBSock, "STAT", iTimeout, szResponse, sizeof(szResponse) - 1))
        return (FALSE);

    char           *pszResponse = szResponse;

    for (; (*pszResponse != '\0') && !isdigit(*pszResponse); pszResponse++);

    if (!isdigit(*pszResponse))
        return (FALSE);

    char            szDigits[128] = "";

    for (int ii = 0; isdigit(*pszResponse); ii++, pszResponse++)
        szDigits[ii] = *pszResponse;

    szDigits[ii] = '\0';

    lNumMessages = atol(szDigits);

    for (; (*pszResponse != '\0') && !isdigit(*pszResponse); pszResponse++);

    if (!isdigit(*pszResponse))
        return (FALSE);

    for (ii = 0; isdigit(*pszResponse); ii++, pszResponse++)
        szDigits[ii] = *pszResponse;

    szDigits[ii] = '\0';

    lBytesCount = atol(szDigits);

    return (TRUE);

}





//
// Funzione GM_EXECRESCOMMAND
//
// @func static BOOL | GM_ExecResCommand | DESCRIZIONE
//
// @parm BSOCK_HANDLE | hBSock | DESCRIZIONE
// @parm const char | *pszRespFileName | DESCRIZIONE
// @parm const char | *pszCommand | DESCRIZIONE
// @parm int | iTimeout | DESCRIZIONE
//
// @rdesc DESCRIZIONE_VALORE_RITORNO
//
// MadeBy : Davide Libenzi
//
static BOOL     GM_ExecResCommand(BSOCK_HANDLE hBSock, const char *pszRespFileName,
                        const char *pszCommand, int iTimeout)
{

    char            szRespBuffer[RECV_BUFFER_SIZE] = "";

    if (!GM_ExecCommand(hBSock, pszCommand, iTimeout, szRespBuffer, sizeof(szRespBuffer) - 1))
        return (FALSE);

    FILE           *pResFile = fopen(pszRespFileName, "wb");

    if (pResFile == NULL)
        perror(pszRespFileName);

    while (BSckGetString(hBSock, szRespBuffer, sizeof(szRespBuffer) - 1, GM_TIMEOUT) != NULL)
    {
        if (strcmp(szRespBuffer, ".") == 0)
            break;

        if (szRespBuffer[0] == '.')
            for (int ii = 0; szRespBuffer[ii] != '\0'; ii++)
                szRespBuffer[ii] = szRespBuffer[ii + 1];

        if (pResFile != NULL)
        {
            fputs(szRespBuffer, pResFile);
            fputs("\r\n", pResFile);
        }
    }

    if (pResFile != NULL)
        fclose(pResFile);

    return (TRUE);

}




int             main(int argc, char *argv[])
{

    int             iPort = 110,
                    iTimeout = GM_TIMEOUT;

    if (argc < 4)
    {
        printf("usage : %s  server  user  password [-d] [-p port {%d}] [-t timeout {%d}]\n",
                argv[0], iPort, iTimeout);
        return (1);
    }

    BOOL            bRemoveMessages = FALSE;
    char           *pszServer = argv[1],
                   *pszUserName = argv[2],
                   *pszPassword = argv[3];

    for (int ii = 4; ii < argc; ii++)
    {
        if (argv[ii][0] == '-')
        {
            switch (argv[ii][1])
            {
                case ('p'):
                    if ((++ii < argc) && (argv[ii][0] != '-'))
                        iPort = atoi(argv[ii]);
                    break;

                case ('d'):
                    bRemoveMessages = TRUE;
                    break;

                case ('t'):
                    if ((++ii < argc) && (argv[ii][0] != '-'))
                        iTimeout = atoi(argv[ii]);
                    break;
            }
        }
    }


    char            szSockDesc[256] = "";

    if (!SvrInitSockets(szSockDesc))
    {
        printf("network library error\n");
        return (2);
    }

    BSOCK_HANDLE    hBSock = GM_Login(pszServer, iPort, pszUserName, pszPassword, iTimeout);

    if (hBSock == INVALID_BSOCK_HANDLE)
    {
        printf("cannot login \"%s\":\"%s\" to \"%s\" on port %d\n",
                pszUserName, pszPassword, pszServer, iPort);
        SvrCleanupSockets();
        return (3);
    }

    long            lNumMessages = 0,
                    lBytesCount = 0;

    if (!GM_DoStat(hBSock, lNumMessages, lBytesCount, iTimeout))
    {

        GM_Logout(hBSock, iTimeout);
        SvrCleanupSockets();
        return (4);
    }

    printf("[ %ld messages : %ld bytes ] in mailbox\n", lNumMessages, lBytesCount);


    for (int iMessageId = 0; iMessageId < (int) lNumMessages; iMessageId++)
    {
        char            szFileName[MAX_PATH] = "";

        sprintf(szFileName, "%s-%s-%04d.eml", pszServer, pszUserName, iMessageId + 1);


        printf("message %03d ... ", iMessageId + 1);


        char            szCommand[256] = "";

        sprintf(szCommand, "RETR %d", iMessageId + 1);

        if (GM_ExecResCommand(hBSock, szFileName, szCommand, iTimeout))
        {
            printf("OK\n");

            if (bRemoveMessages)
            {
                printf("removing %03d ... ", iMessageId + 1);


                sprintf(szCommand, "DELE %d", iMessageId + 1);

                if (GM_ExecCommand(hBSock, szCommand, iTimeout))
                    printf("OK\n");
                else
                {
                    printf("Fail\n");

                    break;
                }
            }
        }
        else
        {
            printf("Fail\n");

            remove(szFileName);
            break;
        }
    }


    GM_Logout(hBSock, iTimeout);

    SvrCleanupSockets();

    return (0);

}
